CREATE OR REPLACE PACKAGE BODY xpath AS

  -- "Casts" an DOMDocument as a DOMNode
  FUNCTION toNode(doc xmldom.DOMDocument) RETURN xmldom.DOMNode IS
  BEGIN RETURN xmldom.makeNode(doc); END;

  ----------------------------------------------------------------------
  -- Remove extraneous whitespace from a string
  -- Translates CR, LF, and TAB to spaces then leaves only a single
  -- space between characters.
  ----------------------------------------------------------------------
  FUNCTION normalizeWS( v VARCHAR2 )
  RETURN VARCHAR2 IS
    result VARCHAR2(32767);
  BEGIN
    result := RTRIM(LTRIM(TRANSLATE(v,CHR(13)||CHR(8)||CHR(10),'   ')));
    WHILE (INSTR(result,'  ') > 0) LOOP
      result := REPLACE(result,'  ',' ');
    END LOOP;
    RETURN result;
  END;

  ----------------------------------------------------------------------
  -- Selects nodes matching an XPath expression and returns the string
  -- consisting of the XML serialization of each matching node in
  -- document order. Passing an XPath of '/' matches the single root node
  -- and so the entire document is serialized to XML markup. If the
  -- normalizeWhitespace is set to TRUE (the default) then the string
  -- is stripped of extraneous whitespace before being returned.
  ----------------------------------------------------------------------
  FUNCTION selectAndPrint(doc xmldom.DOMDocument,xpath VARCHAR2,
                          normalizeWhitespace BOOLEAN := TRUE)
  RETURN VARCHAR2 IS
    retval   VARCHAR2(32767);
    result   VARCHAR2(32767);
    curNode  xmldom.DOMNode;
    nodeType NATURAL;
    matches  xmldom.DOMNodeList;
  BEGIN
    IF xpath = '/' THEN
      xmldom.writeToBuffer(doc,retval);
    ELSE
      matches  := xslprocessor.selectNodes(toNode(doc),xpath);
      FOR i IN 1..xmldom.getLength(matches) LOOP
        curNode := xmldom.item(matches,i-1);
        xmldom.writeToBuffer(curNode,result);
        nodeType := xmldom.getNodeType(curNode);
        IF nodeType NOT IN (xmldom.TEXT_NODE,xmldom.CDATA_SECTION_NODE) THEN
          retval := retval||RTRIM(RTRIM(result,CHR(10)),CHR(13));
        ELSE
          retval := retval||result;
        END IF;
      END LOOP;
    END IF;
    IF normalizeWhitespace THEN
      RETURN normalizeWS(retval);
    ELSE
      RETURN retval;
    END IF;
  END;

  ----------------------------------------------------------------------
  -- The "extract" functions parse the inbound document if necessary
  -- and use selectAndPrint to extract the XML markup for the nodes
  -- matching the XPath expression passed in. If the normalize parameter
  -- is TRUE, extraneous whitespace is removed.
  ----------------------------------------------------------------------
  FUNCTION extract(doc xmldom.DOMDocument,xpath VARCHAR2:='/',
                   normalize BOOLEAN:=TRUE)
    RETURN VARCHAR2 IS
  BEGIN
    IF xmldom.isNull(doc) OR xpath IS NULL THEN RETURN NULL; END IF;
    RETURN selectAndPrint(doc,xpath,normalize);
  END;

  FUNCTION extract(doc VARCHAR2,xpath VARCHAR2 := '/', normalize BOOLEAN:=TRUE)
    RETURN VARCHAR2 IS
    xmldoc xmldom.DOMDocument;
    retval VARCHAR2(32767);
  BEGIN
    IF doc IS NULL OR xpath IS NULL THEN RETURN NULL; END IF;
    xmldoc := xml.parse(doc);
    retval := selectAndPrint(xmldoc,xpath,normalize);
    xml.freeDocument(xmldoc);
    RETURN retval;
  EXCEPTION
    WHEN OTHERS THEN xml.freeDocument(xmldoc); RAISE;
  END;

  FUNCTION extract(doc CLOB,xpath VARCHAR2 := '/',normalize BOOLEAN:=TRUE)
   RETURN VARCHAR2 IS
    xmldoc xmldom.DOMDocument;
    retval VARCHAR2(32767);
  BEGIN
    IF doc IS NULL OR xpath IS NULL THEN RETURN NULL; END IF;
    xmldoc := xml.parse(doc);
    retval := selectAndPrint(xmldoc,xpath,normalize);
    xml.freeDocument(xmldoc);
    RETURN retval;
  EXCEPTION
    WHEN OTHERS THEN xml.freeDocument(xmldoc); RAISE;
  END;

  ----------------------------------------------------------------------
  -- The "valueOf" functions parse the inbound document if necessary
  -- and use xslprocessor.valueOf to return the value of the matching
  -- expression. The semantics of valueOf are the same as <xsl:value-of>
  -- The XPath expression is interpreted with respect to the root
  -- element of the document, except if valueOf(node,...) is used in
  -- which case the expression is interpreted with resped to the
  -- current node.
  ----------------------------------------------------------------------
  FUNCTION valueOf(node xmldom.DOMNode, xpath VARCHAR2,normalize BOOLEAN:=FALSE)
    RETURN VARCHAR2 IS
  BEGIN
    IF xmldom.IsNull(node) OR xpath IS NULL THEN RETURN NULL; END IF;
    IF normalize THEN
      RETURN normalizeWS(xslprocessor.valueOf(node,xpath));
    ELSE
      RETURN xslprocessor.valueOf(node,xpath);
    END IF;
  END;

  FUNCTION valueOf(doc xmldom.DOMDocument,xpath VARCHAR2,normalize BOOLEAN:=FALSE)
    RETURN VARCHAR2 IS
  BEGIN
    IF xmldom.IsNull(doc) OR xpath IS NULL THEN RETURN NULL; END IF;
    RETURN valueOf(toNode(doc),xpath,normalize);
  END;

  FUNCTION valueOf(doc VARCHAR2,xpath VARCHAR2,normalize BOOLEAN:=FALSE)
    RETURN VARCHAR2 IS
    xmldoc xmldom.DOMDocument;
    retval VARCHAR2(32767);
  BEGIN
    IF doc IS NULL OR xpath IS NULL THEN RETURN NULL; END IF;
    xmldoc := xml.parse(doc);
    retval := valueOf(xmldoc,xpath,normalize);
    xml.freeDocument(xmldoc);
    RETURN retval;
  EXCEPTION
    WHEN OTHERS THEN xml.freeDocument(xmldoc); RAISE;
  END;

  FUNCTION valueOf(doc CLOB,xpath VARCHAR2,normalize BOOLEAN:=FALSE)
    RETURN VARCHAR2 IS
    xmldoc xmldom.DOMDocument;
    retval VARCHAR2(32767);
  BEGIN
    IF doc IS NULL OR xpath IS NULL THEN RETURN NULL; END IF;
    xmldoc := xml.parse(doc);
    retval := valueOf(xmldoc,xpath,normalize);
    xml.freeDocument(xmldoc);
    RETURN retval;
  EXCEPTION
    WHEN OTHERS THEN xml.freeDocument(xmldoc); RAISE;
  END;

  ----------------------------------------------------------------------
  -- The "selectNodes" functions parse the inbound document if necessary
  -- and use xslprocessor.selectNodes to return a DOM NodeList of
  -- nodes matching the XPath expression. To iterate over the node list
  -- define "curNode" to be of type xmldom.DOMNode and write a loop like:
  --
  --   FOR n IN 1..xmldom.getLength(yourNodeList) LOOP
  --     curNode := xmldom.item(yourNodeList, n - 1 );
  --        :
  --   END LOOP;
  ----------------------------------------------------------------------
  FUNCTION  selectNodes(node xmldom.DOMNode,xpath VARCHAR2)
  RETURN xmldom.DOMNodeList IS
  BEGIN
    RETURN xslprocessor.selectNodes(node,xpath);
  END;

  FUNCTION  selectNodes(doc xmldom.DOMDocument,xpath VARCHAR2)
  RETURN xmldom.DOMNodeList IS
  BEGIN
    RETURN selectNodes(toNode(doc),xpath);
  END;

  FUNCTION  selectNodes(doc VARCHAR2, xpath VARCHAR2)
  RETURN xmldom.DOMNodeList IS
    xmldoc xmldom.DOMDocument;
    retlist xmldom.DOMNodeList;
  BEGIN
    xmldoc := xml.parse(doc);
    retlist := xslprocessor.selectNodes(toNode(xmldoc),xpath);
    xml.freeDocument(xmldoc);
    RETURN retlist;
  EXCEPTION
    WHEN OTHERS THEN xml.freeDocument(xmldoc); RAISE;
  END;

  FUNCTION  selectNodes(doc CLOB, xpath VARCHAR2)
  RETURN xmldom.DOMNodeList IS
    xmldoc xmldom.DOMDocument;
    retlist xmldom.DOMNodeList;
  BEGIN
    xmldoc := xml.parse(doc);
    retlist := xslprocessor.selectNodes(toNode(xmldoc),xpath);
    xml.freeDocument(xmldoc);
    RETURN retlist;
  EXCEPTION
    WHEN OTHERS THEN xml.freeDocument(xmldoc); RAISE;
  END;

  ----------------------------------------------------------------------
  -- The "test" functions parse the inbound document if necessary
  -- and use selectNodes above to see if XPath predicate is true
  -- with respect to the current node
  ----------------------------------------------------------------------
  FUNCTION test(doc xmldom.DOMDocument,xpath VARCHAR2) RETURN BOOLEAN IS
  BEGIN
    RETURN xmldom.getLength(selectNodes(doc,'/self::node()['||xpath||']')) > 0;
  END;

  FUNCTION test(node xmldom.DOMNode,xpath VARCHAR2) RETURN BOOLEAN IS
  BEGIN
    RETURN xmldom.getLength(selectNodes(node,'./self::node()['||xpath||']')) > 0;
  END;

  FUNCTION test(doc VARCHAR2, xpath VARCHAR2) RETURN BOOLEAN IS
    xmldoc xmldom.DOMDocument;
    retval BOOLEAN;
  BEGIN
    xmldoc := xml.parse(doc);
    retval := test(xmldoc,xpath);
    xml.freeDocument(xmldoc);
    RETURN retval;
  EXCEPTION
    WHEN OTHERS THEN xml.freeDocument(xmldoc); RAISE;
  END;

  FUNCTION test(doc CLOB, xpath VARCHAR2) RETURN BOOLEAN IS
    xmldoc xmldom.DOMDocument;
    retval BOOLEAN;
  BEGIN
    xmldoc := xml.parse(doc);
    retval := test(xmldoc,xpath);
    xml.freeDocument(xmldoc);
    RETURN retval;
  EXCEPTION
    WHEN OTHERS THEN xml.freeDocument(xmldoc); RAISE;
  END;
  
END;

