CREATE OR REPLACE PACKAGE BODY xslt AS

  ----------------------------------------------------------------------
  -- Reset the parameters on the stylesheet and then set the new
  -- parameter values based on the parameters in the paramlist
  ----------------------------------------------------------------------
  PROCEDURE setParams(style xslprocessor.Stylesheet, params paramlist) IS
  BEGIN
    IF params.exists(1) THEN
      xslprocessor.resetParams(style);
      FOR i IN 1..params.COUNT LOOP
        IF params(i).name IS NOT NULL AND params(i).value IS NOT NULL THEN
          xslprocessor.setParam(style,params(i).name,'"'||params(i).value||'"');
        END IF;
      END LOOP;
    END IF;
  END;

  ----------------------------------------------------------------------
  -- The "stylesheet" functions create a new stylesheet object from
  -- an XML source. Remember to call freeStylesheet() when you're done
  -- using the stylesheet.
  ----------------------------------------------------------------------
  FUNCTION stylesheet(doc xmldom.DOMDocument) RETURN xslprocessor.Stylesheet IS
  BEGIN
    return xslprocessor.newStylesheet(doc,NULL);
  END;

  FUNCTION stylesheet(doc VARCHAR2) RETURN xslprocessor.Stylesheet IS
    xmldoc xmldom.DOMDocument;
    newsheet xslprocessor.Stylesheet;
  BEGIN
    xmldoc   := xml.parse(doc);
    newsheet := xslprocessor.newStylesheet(xmldoc,NULL);
    xml.freeDocument(xmldoc);
    RETURN newsheet;
  EXCEPTION
    WHEN OTHERS THEN xml.freeDocument(xmldoc); RAISE;
  END;

  FUNCTION stylesheetFromURL(url VARCHAR2) RETURN xslprocessor.Stylesheet IS
  BEGIN
    return xslprocessor.newStylesheet(url,url);
  END;

  FUNCTION stylesheet(doc CLOB) RETURN xslprocessor.Stylesheet IS
    xmldoc xmldom.DOMDocument;
    newsheet xslprocessor.Stylesheet;
  BEGIN
    xmldoc := xml.parse(doc);
    newsheet := xslprocessor.newStylesheet(xmldoc,NULL);
    xml.freeDocument(xmldoc);
    RETURN newsheet;
  EXCEPTION
    WHEN OTHERS THEN xml.freeDocument(xmldoc); RAISE;
  END;

  FUNCTION stylesheet(doc BFILE) RETURN xslprocessor.Stylesheet IS
    xmldoc xmldom.DOMDocument;
    newsheet xslprocessor.Stylesheet;
  BEGIN
    xmldoc := xml.parse(doc);
    newsheet := xslprocessor.newStylesheet(xmldoc,NULL);
    xml.freeDocument(xmldoc);
    RETURN newsheet;
  EXCEPTION
    WHEN OTHERS THEN xml.freeDocument(xmldoc); RAISE;
  END;

  ----------------------------------------------------------------------
  -- The "transform" functions carry out the transformation described
  -- by the XSLT stylesheet object passed-in, on the XML document source
  -- The resulting target document is returned as a String.
  ----------------------------------------------------------------------
  FUNCTION transform(source xmldom.DOMDocument,
                     style  xslprocessor.Stylesheet,
                     params paramlist := none) RETURN VARCHAR2 IS
    engine xslprocessor.Processor := xslprocessor.newProcessor;
    out VARCHAR2(32767);
  BEGIN
    setParams(style,params);
    xslprocessor.processXSL(engine,style,source,out);
    xslprocessor.freeProcessor(engine);
    RETURN out;
  EXCEPTION
    WHEN OTHERS THEN xslprocessor.freeProcessor(engine); RAISE;
  END;

  FUNCTION transform(source VARCHAR2,
                     style  xslprocessor.Stylesheet,
                     params paramlist := none) RETURN VARCHAR2 IS
    xmldoc xmldom.DOMDocument;
    retval VARCHAR2(32767);
  BEGIN
    xmldoc := xml.parse(source);
    retval := transform(xmldoc,style,params);
    xml.freeDocument(xmldoc);
    RETURN retval;
  EXCEPTION
    WHEN OTHERS THEN xml.freeDocument(xmldoc); RAISE;
  END;

  FUNCTION transform(source CLOB,
                     style  xslprocessor.Stylesheet,
                     params paramlist := none) RETURN VARCHAR2 IS
    xmldoc xmldom.DOMDocument;
    retval VARCHAR2(32767);
  BEGIN
    xmldoc := xml.parse(source);
    retval := transform(xmldoc,style,params);
    xml.freeDocument(xmldoc);
    RETURN retval;
  EXCEPTION
    WHEN OTHERS THEN xml.freeDocument(xmldoc); RAISE;
  END;

  ----------------------------------------------------------------------
  -- The "transformToDOM" functions carry out the transformation
  -- described by the XSLT stylesheet object passed-in on the XML
  -- document source. The resulting target document is returned as an
  -- in-memory XML Document for further processing. Using this function
  -- you can "chain" stylesheets together by making the result of one
  -- transformation be the source of the next.
  ----------------------------------------------------------------------
  FUNCTION transformToDOM(source xmldom.DOMDocument,
                          style  xslprocessor.Stylesheet,
                          params paramlist := none) RETURN xmldom.DOMDocument IS
    engine xslprocessor.Processor := xslprocessor.newProcessor;
    root   xmldom.DOMNode         := xmldom.makeNode(xmldom.newDOMDocument);
    result xmldom.DOMNode;
  BEGIN
    setParams(style,params);
    result := xmldom.makeNode(xslprocessor.processXSL(engine,style,source));
    RETURN xmldom.makeDocument(xmldom.appendChild(root,result));
    xslprocessor.freeProcessor(engine);
  EXCEPTION
    WHEN OTHERS THEN xslprocessor.freeProcessor(engine); RAISE;
  END;

  FUNCTION transformToDOM(source VARCHAR2,
                          style  xslprocessor.Stylesheet,
                          params paramlist := none) RETURN xmldom.DOMDocument IS
    xmldoc xmldom.DOMDocument;
    retdoc xmldom.DOMDocument;
  BEGIN
    xmldoc := xml.parse(source);
    retdoc := transformToDOM(xmldoc,style,params);
    xml.freeDocument(xmldoc);
    RETURN retdoc;
  EXCEPTION
    WHEN OTHERS THEN xml.freeDocument(xmldoc); RAISE;
  END;

  FUNCTION transformToDOM(source CLOB,
                          style  xslprocessor.Stylesheet,
                          params paramlist := none) RETURN xmldom.DOMDocument IS
    xmldoc xmldom.DOMDocument;
    retdoc xmldom.DOMDocument;
  BEGIN
    xmldoc := xml.parse(source);
    retdoc := transformToDOM(xmldoc,style,params);
    xml.freeDocument(xmldoc);
    RETURN retdoc;
  EXCEPTION
    WHEN OTHERS THEN xml.freeDocument(xmldoc); RAISE;
  END;

  ----------------------------------------------------------------------
  -- Construct a paramlist object based on one or more pairs
  -- of 'name','value' parameters.
  ----------------------------------------------------------------------
  FUNCTION params( n1 VARCHAR2,      v1 VARCHAR2,
                   n2 VARCHAR2:=NULL,v2 VARCHAR2:=NULL,
                   n3 VARCHAR2:=NULL,v3 VARCHAR2:=NULL,
                   n4 VARCHAR2:=NULL,v4 VARCHAR2:=NULL,
                   n5 VARCHAR2:=NULL,v5 VARCHAR2:=NULL) RETURN paramlist IS
    p_list paramlist;
  BEGIN
    p_list(1).name := n1; p_list(1).value := v1;
    p_list(2).name := n2; p_list(2).value := v2;
    p_list(3).name := n2; p_list(3).value := v2;
    p_list(4).name := n2; p_list(4).value := v2;
    p_list(5).name := n2; p_list(5).value := v2;
    RETURN p_list;
  END;

  PROCEDURE freeStylesheet( style xslprocessor.Stylesheet) IS
  BEGIN
    xslprocessor.freeStylesheet(style);
  END;

END;
